
#include "theme.h"
#include "themeCursor.h"
#include "themeLabel.h"

char *safeStrdup (char *str) {
	char *res = strdup (str);

	if (res == NULL) {
		perror ("unable to duplicate string... exiting");
		exit (1);
	}
	return res;
}

void ThemeSendExposeToSelfAndChildren (Display *dis, Window win) {
	Window root;
	Window *children;
	Window parent;
	unsigned int nchildren;
	unsigned int i;
	XEvent event;
	XWindowAttributes xattr;
	
	event.type = Expose;
	event.xexpose.type = Expose;
	event.xexpose.send_event = 0;
	event.xexpose.display = dis;
	event.xexpose.window = win;
	
	XGetWindowAttributes (dis, win, &xattr);
	
	event.xexpose.x = 0;
	event.xexpose.y = 0;
	event.xexpose.width = xattr.width;
	event.xexpose.height = xattr.height;
	
	XSendEvent (dis, win, 1, ExposureMask, &event);
	XClearArea (dis, win, 0, 0, xattr.width, xattr.height, 1);
	
	if (!XQueryTree (dis, win, &root, &parent, &children, &nchildren)) {
		return;
	}
	
	for (i = 0; i < nchildren; ++i) {
		ThemeSendExposeToSelfAndChildren (dis, children[i]);
	}
	XFree (children);
}


typedef struct ThemeFrame {
	Pixmap pix;
	Display *dis;
	Tk_Window tkwin;
	unsigned int depth;
	int screen;
	Window xWin;
	unsigned int width;
	unsigned int height;
	int kdeSystray;
} ThemeFrame;


static void ThemeFrameSetPixmap (ClientData clientData) {
	ThemeFrame *wPtr = (ThemeFrame *) clientData;
	Pixmap tmpPix;
	GC tmpGC;

	if (wPtr->pix == None) {
		Tk_SetWindowBackgroundPixmap (wPtr->tkwin, ParentRelative);
		XClearWindow (wPtr->dis, wPtr->xWin); 
		return;
	}

	tmpGC = XCreateGC (wPtr->dis, wPtr->xWin, 0, 0);
	tmpPix = XCreatePixmap (wPtr->dis, 
		wPtr->xWin, 
		wPtr->width, wPtr->height,
		wPtr->depth
	);
	
	XCopyArea (wPtr->dis, 
		wPtr->pix, tmpPix,
		tmpGC,
		0, 0, 
		wPtr->width, wPtr->height,
		0, 0
	);
	
	Tk_SetWindowBackgroundPixmap (wPtr->tkwin, tmpPix);
	
	XClearWindow (wPtr->dis, wPtr->xWin);
	XFreeGC (wPtr->dis, tmpGC);
	XFreePixmap (wPtr->dis, tmpPix);
}

static void ThemeFrameEventDestroy (ThemeFrame *wPtr) {
	if (wPtr->pix != None) {
		XFreePixmap (wPtr->dis, wPtr->pix);
		ckfree ((char *) wPtr);
	}
}

static void ThemeFrameEventSendExposeWhenIdle (ClientData clientData) {
	ThemeFrame *wPtr = (ThemeFrame *) clientData;
	
	ThemeSendExposeToSelfAndChildren (wPtr->dis, wPtr->xWin);
}

static void ThemeFrameEvent (ClientData clientData, XEvent *eventPtr) {
	ThemeFrame *wPtr = (ThemeFrame *) clientData;
	
	if (eventPtr->type == DestroyNotify) {
		ThemeFrameEventDestroy (wPtr);
		return;
	} 
	
	if (eventPtr->type == ConfigureNotify) {
		Tcl_DoWhenIdle (ThemeFrameSetPixmap, (ClientData) wPtr);
		Tcl_DoWhenIdle (ThemeFrameEventSendExposeWhenIdle, (ClientData) wPtr);
	} else {
		Tcl_DoWhenIdle (ThemeFrameSetPixmap, (ClientData) wPtr);
	}
}

static int ThemeFrameInstanceCmd OBJ_CMD_ARGS {
	ThemeFrame *wPtr = (ThemeFrame *) clientData;
	char *data;
	Pixmap pixShape;
	int ijunk;
	unsigned int uijunk;
	Window wjunk;

	if (objc != 2) {
		Tcl_WrongNumArgs (interp, 1, objv, "xpmData");
		return TCL_ERROR;
	}

	data = Tcl_GetString (objv[1]);
	
	if (wPtr->pix != 0) {
		XFreePixmap (wPtr->dis, wPtr->pix); 
	}

	if (XpmCreatePixmapFromBuffer (wPtr->dis, wPtr->xWin, data, &wPtr->pix, &pixShape, NULL)) {
		Tcl_SetResult (interp, "bad xpm", TCL_STATIC);
		wPtr->pix = None;
		return TCL_ERROR;
	}
	
	if (wPtr->pix == 0) {
		Tcl_SetResult (interp, "invalid xpm pixmap", TCL_STATIC);
		wPtr->pix = None;
		return TCL_ERROR;
	}
	
	XGetGeometry (wPtr->dis, 
		wPtr->pix, 
		&wjunk,
		&ijunk, &ijunk,
		&wPtr->width, &wPtr->height,
		&uijunk,
		&wPtr->depth
	);

	ThemeFrameSetPixmap (wPtr);
	ThemeSendExposeToSelfAndChildren (wPtr->dis, wPtr->xWin);

	return TCL_OK;
}

int ThemeFrameCmd OBJ_CMD_ARGS {
	char *path = NULL;
	Tk_Window tkmain;
	Tk_Window tkwin;
	ThemeFrame *wPtr;
	int i;
	char *class = "ThemeFrame";
	int kdeSystray = 0;
	char *screenName = NULL;
	int width = 1;
	int height = 1;

	/*theme:frame .path*/
	if (objc < 2) {
		Tcl_WrongNumArgs (interp, 1, objv, ".path ?-class aClass?");
		return TCL_ERROR;
	}
	
	path = Tcl_GetString (objv[1]);
	
	for (i = 2; i < objc; ++i) {
		int len = 0;
		char *arg = Tcl_GetStringFromObj (objv[i], &len);
		
		if (strncmp (arg, "-class", len) == 0) {
			++i;
			if (i >= objc) {
				Tcl_SetResult (interp, "give an argument to -class", TCL_STATIC);
				return TCL_ERROR;
			}
			
			class = Tcl_GetString (objv[i]);
		} else if (strncmp (arg, "-kdesystray", len) == 0) {
			screenName = "";
			kdeSystray = 1;

		} else if (strncmp (arg, "-width", len) == 0) {
			++i;
			if (i >= objc) {
				Tcl_SetResult (interp, "give an argument to -width", TCL_STATIC);
				return TCL_ERROR;
			}
			if (Tcl_GetIntFromObj (interp, objv[i], &width) != TCL_OK) {
				return TCL_ERROR;
			}
		} else if (strncmp (arg, "-height", len) == 0) {
			++i;
			if (i >= objc) {
				Tcl_SetResult (interp, "give an argument to -height", TCL_STATIC);
				return TCL_ERROR;
			}
			
			if (Tcl_GetIntFromObj (interp, objv[i], &height) != TCL_OK) {
				return TCL_ERROR;
			}
		} else {
			Tcl_SetResult (interp, "invalid argument", TCL_STATIC);
			return TCL_ERROR;
		}
	}
	
	tkmain = Tk_MainWindow (interp);
	tkwin = Tk_CreateWindowFromPath (interp, tkmain, path, screenName);
	if (tkwin == NULL) {
		return TCL_ERROR;
	}
	
	Tk_MakeWindowExist (tkwin);

	Tk_SetClass (tkwin, class);

	wPtr = (ThemeFrame *) ckalloc (sizeof (ThemeFrame));
	wPtr->width = width;
	wPtr->height = height;
	wPtr->tkwin = tkwin;
	wPtr->xWin = Tk_WindowId (tkwin);
	wPtr->dis = Tk_Display (tkwin);
	wPtr->pix = 0;
	wPtr->screen = DefaultScreen (wPtr->dis);
	wPtr->depth = DefaultDepth (wPtr->dis, wPtr->screen); 
	wPtr->kdeSystray = kdeSystray;
	
	if (kdeSystray) {
		Atom kde_net_wm_system_tray_window_for = XInternAtom (wPtr->dis, "_KDE_NET_WM_SYSTEM_TRAY_WINDOW_FOR", False);
		Window parent;
		Window root;
		Window *children;
		int nchildren = 0;
		
		Tk_SetWindowBorderWidth (tkwin, 0);
		Tk_MapWindow (tkwin);

		/*This finds the real parent that is created by the Tk WM code when Tk_MapWindow is called.*/
		XQueryTree (wPtr->dis, wPtr->xWin, &root, &parent, &children, &nchildren);
		XFree (children);
	
		Tk_UnmapWindow (tkwin);
		XSetWindowBorderWidth (wPtr->dis, parent, 0);

		XChangeProperty (wPtr->dis, parent, 
			kde_net_wm_system_tray_window_for, 
			XA_WINDOW, 32, PropModeReplace,
			(unsigned char *) &parent, 1
		);
		
		Tk_MapWindow (tkwin);
	}
	
	Tcl_CreateObjCommand (interp, path, ThemeFrameInstanceCmd, (ClientData) wPtr, (Tcl_CmdDeleteProc *) NULL);
	
	Tk_CreateEventHandler (tkwin, StructureNotifyMask, ThemeFrameEvent, (ClientData) wPtr);

	Tk_GeometryRequest (tkwin, wPtr->width, wPtr->height);

	Tcl_SetResult (interp, path, TCL_VOLATILE);

	return TCL_OK;
}




/*theme:frame:translucent*/
typedef struct ThemeFrameTranslucent {
	Tk_Window tkwin;
	Window xWin;
	Display *dis;
	int depth;
	Colormap colormap;
	XColor xcol;
} ThemeFrameTranslucent;


int ThemeFrameTranslucentErrorHandler (ClientData cdata, XErrorEvent *errPtr) {
	return 0;
}

static void ThemeFrameTranslucentEventDestroy (ThemeFrameTranslucent *wPtr) {
	XFreeColors (wPtr->dis, wPtr->colormap, &wPtr->xcol.pixel, 1, 0L);
	ckfree ((char *)wPtr);
}

static void ThemeFrameTranslucentDraw (ThemeFrameTranslucent *wPtr) {
	Tk_Window tkParent = Tk_Parent (wPtr->tkwin);
	Window xParent = Tk_WindowId (tkParent);
	GC pGC = XCreateGC (wPtr->dis, wPtr->xWin, 0, 0);
	int imgWidth = Tk_Width (wPtr->tkwin);
	int imgHeight = Tk_Height (wPtr->tkwin);
	int imgX = Tk_X (wPtr->tkwin);
	int imgY = Tk_Y (wPtr->tkwin);
	XImage *img = NULL;
	Pixmap pix = None;
	int numPixels = (imgWidth * imgHeight);
	int pi = 0;
	
	img = XGetImage (wPtr->dis, xParent, imgX, imgY, imgWidth, imgHeight, AllPlanes, ZPixmap);

	if (img == NULL) {
		goto cleanup;
	}

	if (wPtr->depth >= 24) {
		/*We have 8 bits for each r/g/b*/
		Pixel *imgData = (Pixel *) img->data;
		
		for (pi = 0; pi < numPixels; ++pi) {
			u_int32_t tmpPix = imgData[pi];
			u_int32_t newPix;
			int red = (tmpPix >> 16) & 0x0000ff;
			int green = (tmpPix >> 8) & 0x0000ff;
			int blue = (tmpPix & 0x0000ff);
			
			red = (red - (red / 4));
			green = (green - (green / 4));
			blue = (blue - (blue / 4));
			
			newPix = (red << 16) | (green << 8) | blue;
			imgData[pi] = newPix;
		}
	} else if (wPtr->depth >= 15) {
		/*red = (tmpPix & 0xf800) >> 11; green = (tmpPix&0x07c0) >> 6; blue = (tmpPix&0x003e) >> 1?*/
		u_int16_t *imgData = (u_int16_t *) img->data;
		
		for (pi = 0; pi < numPixels; ++pi) {
			u_int16_t tmpPix = imgData[pi];
			u_int16_t newPix;
			int red = (tmpPix & 0xf800) >> 11;
			int green = (tmpPix & 0x07c0) >> 6;
			int blue = (tmpPix & 0x003e) >> 1;

			red = (red - (red / 4));
			green = (green - (green / 4));
			blue = (blue - (blue / 4));
			newPix = (red << 11) | (green << 6) | (blue << 1);
			imgData[pi] = newPix;
		}
	}
	pix = XCreatePixmap (wPtr->dis, wPtr->xWin, imgWidth, imgHeight, wPtr->depth);
	XPutImage (wPtr->dis, pix, pGC, img, 0, 0, 0, 0, imgWidth, imgHeight);
	Tk_SetWindowBackgroundPixmap (wPtr->tkwin, pix);
	XClearWindow (wPtr->dis, wPtr->xWin);

	cleanup:
	if (img != NULL) {
		XDestroyImage (img);
	}
	XFreePixmap (wPtr->dis, pix);
	XFreeGC (wPtr->dis, pGC);
}


static void ThemeFrameTranslucentEvent (ClientData clientData, XEvent *eventPtr) {
	ThemeFrameTranslucent *wPtr = (ThemeFrameTranslucent *) clientData;
	
	if (eventPtr->type == DestroyNotify) {
		ThemeFrameTranslucentEventDestroy (wPtr);
	} else if (eventPtr->type != Expose) {
		ThemeSendExposeToSelfAndChildren (wPtr->dis, wPtr->xWin);
	} else if (eventPtr->xexpose.count == 0) {
		ThemeFrameTranslucentDraw (wPtr);
	}
}

int ThemeFrameTranslucentCmd OBJ_CMD_ARGS {
	Tk_Window tkwin;
	Tk_Window tkmain;
	char *path = NULL;
	char *colName = NULL;
	ThemeFrameTranslucent *wPtr;
	
	if (objc != 3) {
		Tcl_WrongNumArgs (interp, 1, objv, ".path stippleColor");
		return TCL_ERROR;
	}
	
	path = Tcl_GetString (objv[1]);
	colName = Tcl_GetString (objv[2]);
	
	tkmain = Tk_MainWindow (interp);
	tkwin = Tk_CreateWindowFromPath (interp, tkmain, path, (char *) NULL);
	
	if (tkwin == NULL) {
		return TCL_ERROR;
	}
	
	Tk_MakeWindowExist (tkwin);

	wPtr = (ThemeFrameTranslucent *) ckalloc (sizeof (ThemeFrameTranslucent));
	
	wPtr->tkwin = tkwin;
	wPtr->xWin = Tk_WindowId (tkwin);
	wPtr->dis = Tk_Display (tkwin);
	wPtr->depth = Tk_Depth (tkwin);
	wPtr->colormap = DefaultColormap (wPtr->dis, Tk_ScreenNumber (tkwin));

	if (XParseColor (wPtr->dis, wPtr->colormap, colName, &wPtr->xcol) == 0) {
		Tcl_SetResult (interp, "invalid color", TCL_STATIC);
		return TCL_ERROR;
	}
	XAllocColor (wPtr->dis, wPtr->colormap, &wPtr->xcol);
	
	Tk_CreateErrorHandler (wPtr->dis, -1, -1, -1, ThemeFrameTranslucentErrorHandler, (ClientData) NULL);

	/*Tcl_CreateObjCommand (interp, path, ThemeFrameInstanceCmd, (ClientData) wPtr, (Tcl_CmdDeleteProc *) NULL);*/
	Tk_CreateEventHandler (tkwin, ExposureMask | StructureNotifyMask, ThemeFrameTranslucentEvent, (ClientData) wPtr);
	
	/*Tk_GeometryRequest (tkwin, 1, 1);*/
	
	Tcl_SetResult (interp, path, TCL_VOLATILE);
	
	return TCL_OK;
}

int Theme_Init (Tcl_Interp *interp) {
#ifdef USE_STUBS
	Tcl_InitStubs (interp, "8.3", 0);
	Tk_InitStubs (interp, "8.3", 0);
#endif

#define OBJ_CMD(name,func) Tcl_CreateObjCommand(interp, name, func, (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL)
	OBJ_CMD ("theme:frame", ThemeFrameCmd);
	OBJ_CMD ("theme:label", ThemeLabelCmd);
	OBJ_CMD ("theme:cursor", ThemeCursorCmd);
	OBJ_CMD ("theme:frame:translucent", ThemeFrameTranslucentCmd);
	Tcl_PkgProvide (interp, "Tk_Theme", "23");
	
	return TCL_OK;
#undef OBJ_CMD
}
